#-----------------------------------------------------------------------------
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Name:           phlatDat.rb
# Version:        001
# Description:    Writes selected geometry to *.dat format.
#                 Reads *.dat to geometry.
#                 This particular format is for foamcutters.
# Parameters:     None
# Usage:          Select geometry, Plugins/PhlatToDAT... - export as *.dat
#                 Or Plugins/DATtoPhlat... Open *.dat file ...
# Menu Item :     Plugins/PhlatToDAT... or Plugins/DATtoPhlat...
# Context Menu:   None
# Type:           Plugin
# Date:           2010-12-22 17:58 GMT
# Author:         Neil Gillies  neil@iGull.co.uk
# Needs:          datExporter.rb to be in Plugins folder
#                 Also needs cursor_DAT_begin.png & cursor_DAT_end.png cursor files
#                 to be there too.
#                 Works on Mac OSX 10.6.5 SU 7, needs to be checked on PC.
#-----------------------------------------------------------------------------

#                 load"phlatDat.rb"
# Current issues...
# Seems that there are more than a single variation of the foamcutting *.dat format (surprise!)
# The format used by this plugin to export a *.dat file is ...
# xCoOrd SPACE yCoOrd LF
# Data runs from 0,0 to max 1,1 back to 0,0 [it can be any starting point, but must be sequential].

# Sometimes SU will enable duplicate points to happen when grabbing data from a face :-)
# This version seems to be fine v1.004b moved to v1.001.

require 'sketchup.rb'

if( not file_loaded?("phlatDat.rb") )
	add_separator_to_menu("Plugins")
	UI.menu("Plugins").add_item("PhlatToDAT...") { phlatToDat }
	UI.menu("Plugins").add_item("DATtoPhlat...") { datToPhlat
		Sketchup.active_model.select_tool DATtool.new
	}
end

file_loaded("phlatDat.rb")

#UI.messagebox("phlatDat v1.001 loaded " + "\n" + "© N D Gillies 2010")

#---------- PhlatToDAT -------------------------------------------------------------------

def phlatToDat
	model = Sketchup.active_model
	entities = model.active_entities
	theSelection = Sketchup.active_model.selection
	
	if theSelection.count > 1
		UI.messagebox"Too many items selected, select a face only"
		else
		# Search entities in the selection.
		faceCount=0
		theSelection.each { |entity| 
			if entity.typename == "Face"
				faceCount = faceCount + 1
			end
		}
		case faceCount
			when 0
		    UI.messagebox"No faces selected"
			when 1
			# We have a single face...
			createDAT(theSelection)
		end
	end
	
end #phlatToDat

def createDAT(aSelection)
	
	entity=aSelection.first
	datVertices = entity.vertices #edges
	$datTable=Array.new(datVertices.size)
	$datTable.map! { Array.new(2) }
	
	# Put datVertices into $datTable first as
	# SU works internally in inches ...
	
	for i in 0...(datVertices.size) do
		$datTable[i][0]= datVertices[i].position.x.to_l
		$datTable[i][1]= datVertices[i].position.y.to_l
    end
		
		# Remove any nil entries
		$datTable.compact!
		
		@minX=999999
		@minY=999999
		@maxX=-999999
		@maxY=-999999
		
		# Find x & y offsets in SU space...
		for i in 0...($datTable.size) do
			if @minX>$datTable[i][0]
				@minX=$datTable[i][0]
			end
			if @minY>$datTable[i][1]
				@minY=$datTable[i][1]
			end
		end
			
			# Offset the data to zero
			for i in 0...($datTable.size) do
				$datTable[i][0]= ($datTable[i][0]-@minX)
				$datTable[i][1]= ($datTable[i][1]-@minY)
				#puts $datTable[i][0].to_s + "   " + $datTable[i][1].to_s
			end
				
				# Find x & y offsets in SU space...
				for i in 0...($datTable.size) do
					if @maxX<$datTable[i][0]
						@maxX=$datTable[i][0]
					end
					if @maxY<$datTable[i][1]
						@maxY=$datTable[i][1]
					end
				end
					# Get range (used by x&y)
					if @maxX>@maxY
						range=@maxX
						else
						range=@maxY
					end
					
					for i in 0...($datTable.size) do
						$datTable[i][0]= $datTable[i][0]*(1/range)
						$datTable[i][1]= $datTable[i][1]*(1/range)
						#puts $datTable[i][0].to_s + "   " + $datTable[i][1].to_s
					end
						
						# Need to check that there are no 'real' duplicate points in the output file
						# Looks like this is _maybe_ a SU bug ?
						$datTable.uniq! # Remove any duplicate array entries - love this ruby !
						
						saveDAT
						
					end #createDAT(aSelection)
					
					def saveDAT
					$b = UI.savepanel "Save *.DAT File","","phlatToDatStuff.dat"
					if $b != nil
						f = File.new($b, "w")
						
						for i in 0...($datTable.size) do
							f.puts $datTable[i][0].to_s+" "+$datTable[i][1].to_s + "\n"
						end
							f.close
						end
					end
					
					#---------- DATtoPhlat -------------------------------------------------------------------
					
					def datToPhlat
					
					b = UI.openpanel "Open DAT","","*.dat"# Note that the *.dat mask only works on PC, not Mac - shows _all_ the file types.
					f = File.open(b, 'rb')
					$datText = IO.read(b)
					f.close
					
					# Cleanup first then to array...
					$datText=cleanUp($datText)
					$datText=$datText.split(/\n/)
					
					# Quick check to see if first line contains text
					if ($datText[0] =~/[a-zA-Z]/) # Line contains text, get rid
						$datText.delete_at(0)
					end
					
					# Quick check to see if first line contains text again
					if ($datText[0] =~/[a-zA-Z]/) # Line contains text, get rid
						$datText.delete_at(0)
					end
					
					# Remove any nil entries
					$datText.compact!
					
					# Array holding loaded data
					$datArray=Array.new($datText.length)
					$datArray.map! { Array.new(3) }
					
					for i in 0...($datText.length) do
						splitter=$datText[i].split(" ")
						$datArray[i][0]=splitter[0].to_f
						$datArray[i][1]=splitter[1].to_f
						$datArray[i][2]=0.00
						#puts "x="+$datArray[i][0].to_s+"y="+$datArray[i][1].to_s
					end
						# Remove any nil entries
						$datArray.compact! # Ruby again, wonderful !
						
						# New array to hold scaled data
						$datArrayScaled=Array.new($datArray.length)
						$datArrayScaled.map! { Array.new(3) } # Needs a z element to draw the final dat
						$datName=File.basename(b, ".dat")
						#UI.messagebox(File.basename(b, ".dat")+" loaded for DAT tool")
						Sketchup.active_model.select_tool DATtool.new
						
					end #datToPhlat
					
					def cleanUp(theText)
					theText.squeeze!(" ") # Remove multiple spaces
					theText.gsub!("\n ","\n")# Remove leading spaces
					return theText
				end
				
				#-----------------------------------------------------------------------------
				
				class DATtool
				
				def initialize
				@ip1 = nil
				@ip2 = nil
				@xdown = 0
				@ydown = 0
				
				# Cursor data
				cursor_path = Sketchup.find_support_file("cursor_DAT_begin.png", "Plugins/")
				if cursor_path
					@beginCursor = UI.create_cursor(cursor_path, 2, 15)
				end
				
				cursor_path = Sketchup.find_support_file("cursor_DAT_end.png", "Plugins/")
				if cursor_path
					@endCursor = UI.create_cursor(cursor_path, 28, 15)
				end
				UI.set_cursor(@beginCursor)
				
			end
			
			def activate
			# Two points for the DAT endpoints.
			@ip1 = Sketchup::InputPoint.new
			@ip2 = Sketchup::InputPoint.new
			@ip = Sketchup::InputPoint.new
			@drawn = false
			
			# VCB Label
			result = Sketchup.set_status_text "Width", SB_VCB_LABEL
			self.reset(nil)
		end
		
		def deactivate(view)
		view.invalidate if @drawn
	end
	
	def onMouseMove(flags, x, y, view)
		if( @state == 0 )
			# Left edge of DAT.
			@ip.pick view, x, y
			if( @ip != @ip1 )
				# if the point has changed from the last one we got, then
				# see if we need to display the point.  We need to display it
				# if it has a display representation or if the previous point
				# was displayed.  The invalidate method on the view is used
				# to tell the view that something has changed so that you need
				# to refresh the view.
				view.invalidate if( @ip.display? or @ip1.display? )
				@ip1.copy! @ip
				
				# Set tooltip
				view.tooltip = @ip1.tooltip
				UI.set_cursor(@beginCursor)
			end
			else
			# Trailing edge of DAT
			# If you pass another InputPoint on the pick method of InputPoint
			# it uses that second point to do additional inferencing such as
			# parallel to an axis.
			@ip2.pick view, x, y, @ip1
			view.tooltip = @ip2.tooltip if( @ip2.valid? )
			UI.set_cursor(@endCursor)
			view.invalidate
			
			# Update width displayed in VCB
			if( @ip2.valid? )
				length = @ip1.position.distance(@ip2.position)
				Sketchup::set_status_text length.to_s, SB_VCB_VALUE
			end
			
			# Check to see if the mouse was moved far enough to create an DAT.
			# Create an DAT by dragging or click -> click
			if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
				@dragging = true
			end
		end
	end
	
	def onLButtonDown(flags, x, y, view)
		# First click left edge, second click right edge (or the reverse).
		# Create DAT on second click.
		if( @state == 0 )
			@ip1.pick view, x, y
			if( @ip1.valid? )
				@state = 1
				result = Sketchup.set_status_text "Select left edge for "+$datName, SB_PROMPT
				@xdown = x
				@ydown = y
				UI.set_cursor(@endCursor)
			end
			else
			# Create DAT on second click.
			UI.set_cursor(@beginCursor)
			if( @ip2.valid? )
				self.draw_dat(@ip1.position, @ip2.position,view)
				self.reset(view)
			end
		end
		
		# Clear inference lock (if any)
		view.lock_inference
	end
	
	def onLButtonUp(flags, x, y, view)
		# If we are dragging, create the DAT on the mouse up event
		if( @dragging && @ip2.valid? )
			self.draw_dat(@ip1.position, @ip2.position,view)
			self.reset(view)
		end
	end
	
	# Inference lock if key pressed.
	def onKeyDown(key, repeat, flags, view)
		if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
			@shift_down_time = Time.now
			
			# If inference lock, then unlock
			if( view.inference_locked? )
				# Calling lock_inference with no arguments actually unlocks!
				view.lock_inference
				elsif( @state == 0 && @ip1.valid? )
				view.lock_inference @ip1
				elsif( @state == 1 && @ip2.valid? )
				view.lock_inference @ip2, @ip1
			end
		end
	end
	
	# - - - - - - - - - - - - - - - - -
	#        unlock inference
	# - - - - - - - - - - - - - - - - -
	# If user holds down shift key for more than 0.5 second, then we
	# unlock inference on release.  Otherwise, user presses shift once to 
	# lock and a second time to unlock.
	def onKeyUp(key, repeat, flags, view)
		if( key == CONSTRAIN_MODIFIER_KEY &&
		   view.inference_locked? &&
		   (Time.now - @shift_down_time) > 0.5 )
			view.lock_inference
		end
	end
	
	# - - - - - - - - - - - - - - - - -
	#        DAT via VCB
	# - - - - - - - - - - - - - - - - -
	def onUserText(text, view)
		# Only accept input when state is 1 (i.e. getting the second point)
		return if not @state == 1
		return if not @ip2.valid?
		
		# Mugtrap in case it's not a number ...
		begin
			value = text.to_l
			rescue
			# Error parsing the text
			UI.beep
			puts "Can't convert #{text} to a width"
			value = nil
			Sketchup::set_status_text "", SB_VCB_VALUE
		end
		return if !value
		
		# Compute  direction and second point
		pt1 = @ip1.position
		vec = @ip2.position - pt1
		if( vec.length == 0.0 )
			UI.beep
			return
		end
		vec.length = value
		pt2 = pt1 + vec
		
		# - - - - - - - - - - - - - - - - -
		#              draw DAT
		# - - - - - - - - - - - - - - - - -
		self.draw_dat(pt1, pt2, view)
		self.reset(view)
	end
	
	# The draw method is called whenever the view is refreshed.  It lets the
	# tool draw any temp geometry that it needs.
	def draw(view)
		if( @ip1.valid? )
			if( @ip1.display? )
				@ip1.draw(view)
				@drawn = true
			end
			
			if( @ip2.valid? )
				@ip2.draw(view) if( @ip2.display? )
				
				# The set_color_from_line method determines what color
				# to use to draw a line based on its direction.  For example
				# red, green or blue.
				view.set_color_from_line(@ip1, @ip2)
				self.draw_geometry(@ip1.position, @ip2.position, view)
				@drawn = true
			end
		end
	end
	# - - - - - - - - - - - - - - - - -
	# User hit escape
	# - - - - - - - - - - - - - - - - -
	def onCancel(flag, view)
		self.reset(view)
	end
	
	# The following methods are not directly called from SketchUp.  They are
	# internal methods that are used to support the other methods in this class.
	
	# - - - - - - - - - - - - - - - - -
	# Reset tool
	# - - - - - - - - - - - - - - - - -
	
	def reset(view)
		# This variable keeps track of which point we are currently getting
		@state = 0
		
		# Display status bar prompt
		result = Sketchup.set_status_text "Select left edge for "+$datName, SB_PROMPT
		UI.set_cursor(@beginCursor)
		
		# Clear InputPoints
		@ip1.clear
		@ip2.clear
		
		if( view )
			view.tooltip = nil
			view.invalidate if @drawn
		end
		
		@drawn = false
		@dragging = false
	end
	
	# - - - - - - - - - - - - - - - - -
	#              draw rubber band
	# - - - - - - - - - - - - - - - - -
	def draw_geometry(pt1, pt2, view)
		view.draw_line(pt1, pt2) # Rubberband to show position.
	end
	
	# - - - - - - - - - - - - - - - - -
	#              draw DAT
	# - - - - - - - - - - - - - - - - -
	def draw_dat(pt1, pt2, view)
		if  (pt1.x)!=(pt2.x)
			@scaler=(pt2.x-pt1.x)*100000000 # trying to get away from the 'duplicate points' within 0.001" issue
			#puts @scaler.to_s
			
			for i in 0...($datArray.size) do
				$datArrayScaled[i][0]=($datArray[i][0]*@scaler)+pt1.x#x
				$datArrayScaled[i][1]=($datArray[i][1]*@scaler)+pt1.y#y
				$datArrayScaled[i][2]=($datArray[i][2]*@scaler)+pt1.z#z
				#puts "x="+$datArrayScaled[i][0].to_s+"y="+$datArrayScaled[i][1].to_s+"z="+$datArrayScaled[i][2].to_s
			end
				
				# Scale it back down
				t = Geom::Transformation.scaling pt1,0.00000001
				theDAT = Sketchup.active_model.entities.add_group
				theDAT.entities.add_face $datArrayScaled
				theDAT = theDAT.move! t # Scale it down again
			end
		end
		
	end # class DATtool
	
